/*
 * GLUT CUBE DEMO
 * by Wouter Lievens
 * adapted by Tom Van Cutsem
 *
 * This demo renders a 3D cube, while the camera rotates around it.
 *
 * Controls
 * ========
 * - Use the SPACE BAR to toggle inside/outside facing
 */

/*
 * Include the OpenGL and GLUT headers.
 */
#include <math.h>
#include <gl/glut.h>
#include <gl/gl.h>

/*
 * The infamous mathematical constant.
 */
static const float pi = 3.141592654;

/*
 * When set to 0, the cube will be drawn facing outside.
 * When set to 1, the cube will be drawn facing inside.
 */
static int inside = 0;

/**
 * Draws a vertex of the cube, and sets the color depending on the vertex
 * location.
 *
 * Ie. consider plotting at (1, 0, 0):
 *  we set the color to (R,G,B):
 *   (1 * 1/2 + 1/2, 0 * 1/2 + 1/2, 0 * 1/2 + 1/2) 
 * = (1/2 + 1/2, 1/2, 1/2)
 * = (1, 1/2, 1/2)
 *
 * RGB colors are specified using floats ranging from
 * (0,0,0) = black, to (1,1,1) = white
 */
void drawVertex(float x, float y, float z) {
  static const float factor = 0.50f;
  glColor3f(x * factor + 0.5, y * factor + 0.5, z * factor + 0.5);
  glVertex3f(x, y, z);
}

/**
 * Draws a cube where all planes are facing outside.
 */
void drawCubeFacingOutside() {
  glBegin(GL_QUADS);
  {
    // bottom plane
    drawVertex(-1, -1, -1);
    drawVertex(-1, +1, -1);
    drawVertex(+1, +1, -1);
    drawVertex(+1, -1, -1);
    // top plane
    drawVertex(-1, -1, +1);
    drawVertex(+1, -1, +1);
    drawVertex(+1, +1, +1);
    drawVertex(-1, +1, +1);
    // front plane
    drawVertex(-1, -1, -1);
    drawVertex(+1, -1, -1);
    drawVertex(+1, -1, +1);
    drawVertex(-1, -1, +1);
    // back plane
    drawVertex(-1, +1, -1);
    drawVertex(-1, +1, +1);
    drawVertex(+1, +1, +1);
    drawVertex(+1, +1, -1);
    // left plane
    drawVertex(-1, -1, -1);
    drawVertex(-1, -1, +1);
    drawVertex(-1, +1, +1);
    drawVertex(-1, +1, -1);
    // right plane
    drawVertex(+1, -1, -1);
    drawVertex(+1, +1, -1);
    drawVertex(+1, +1, +1);
    drawVertex(+1, -1, +1);
  }
  glEnd();
}

/**
 * Draws a cube where all planes are facing inside.
 */
void drawCubeFacingInside() {
  glBegin(GL_QUADS);
  {
    // bottom plane
    drawVertex(-1, -1, -1);
    drawVertex(+1, -1, -1);
    drawVertex(+1, +1, -1);
    drawVertex(-1, +1, -1);
    // top plane
    drawVertex(-1, -1, +1);
    drawVertex(-1, +1, +1);
    drawVertex(+1, +1, +1);
    drawVertex(+1, -1, +1);
    // front plane
    drawVertex(-1, -1, -1);
    drawVertex(-1, -1, +1);
    drawVertex(+1, -1, +1);
    drawVertex(+1, -1, -1);
    // back plane
    drawVertex(-1, +1, -1);
    drawVertex(+1, +1, -1);
    drawVertex(+1, +1, +1);
    drawVertex(-1, +1, +1);
    // left plane
    drawVertex(-1, -1, -1);
    drawVertex(-1, +1, -1);
    drawVertex(-1, +1, +1);
    drawVertex(-1, -1, +1);
    // right plane
    drawVertex(+1, -1, -1);
    drawVertex(+1, -1, +1);
    drawVertex(+1, +1, +1);
    drawVertex(+1, +1, -1);
  }
  glEnd();
}

/**
 * Renders the GL scene.
 */
void renderscene() {

  /*
   * Angle variable used to remember the camera's angle, as it orbits the
   * origin.
   */
  static float angle = 0;
  /*
   * The radius at which the camera orbits the origin.
   */
  static const float radius = 4;
  /*
   * The amplitude of the tilting on the Z axis.
   */
  static const float tilt = 2;
  /*
   * The speed factor (relative to the orbital speed) at which tilting occurs.
   */
  static const float tiltspeed = 4;
  
  /*************************
   * Extra motion variables
   *************************/
  static float scalefactor = 1.0; // zooming factor
  static float a = 0; // rotation parameter
  
  // Clears the screen buffer and the depth buffer.
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // You can rotate the cube with this
  // glMatrixMode(GL_MODELVIEW);
  // glRotatef(180,1.0,1.0,1.0);
  
  // Shrinks the cube in half
  // glScalef(0.5,0.5,0.5);
  
  /*********************
   * Zooming in and out
   *********************/
  // first, select modelview matrix to manipulate cube itself
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  
  // zoom in or out
  glScalef(cos(scalefactor), cos(scalefactor), cos(scalefactor));
  scalefactor += 0.01;
  
  // select the projection matrix
  // (controls how the scene is rendered)
  glMatrixMode(GL_PROJECTION);
  // reset the matrix to the identity matrix
  glLoadIdentity();
  
  /**********************************
   * make cube turn counterclockwise
   **********************************/
  a += 0.01;
  glTranslatef(cos(a),sin(a),0);
  
  // load a perspective matrix
  gluPerspective(90,   // the vertical field of view, set to 90 degrees
                 1.0,  // the aspect ratio of width/height
                 0.5,  // the 'near' clipping plane
                 100); // the 'far' clipping plane
  
  // setup the camera
  gluLookAt(cos(angle) * radius, // the x coordinate of the Eye
            sin(angle) * radius, // the y coordinate of the Eye
            cos(tiltspeed * angle) * tilt, // the z coordinate of the Eye
            0, 0, 0, // the x, y, z target that the camera looks at
            0, 0, 1); // the orientation of the world, [0, 0, 1] means 'Z is UP'
  
  // increment the angle variable
  angle += 0.005;
  
  // draw the cube, facing either inside or outside
  if (inside)
    drawCubeFacingInside();
  else
    drawCubeFacingOutside();

  // swap the drawing buffers, effectively drawing the rendered scene to screen
  glutSwapBuffers();
}

/*
 * GLUT callback used to process keyboard input.
 */
void presskey(unsigned char key, int x, int y) {
  // if ESCAPE is pressed, exit the program
  if (key == 27)
  {
    exit(0);
  }
  // if SPACE is pressed, toggle the inside variable
  if (key == ' ')
  {
    inside = !inside;
  }
}

/*
 * The main drawing timer, running at maximal framerate.
 */
void timerfunc(int value) {
  // re-render (ie calls the renderscene function again)
  glutPostRedisplay();
  // re-register the timer
  glutTimerFunc(1, timerfunc, 1);
}

/*
 * Program entry point.
 */
int main(int argc, char *argv[]) {
  // initialise glut
  glutInit(&argc, argv);
  // initialise the display mode
  // - using the depth buffer
  // - using double backbuffering
  // - using RGB color settings
  glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGB);
  // create the window
  glutCreateWindow("Spinning Cube");
  // display it in full screen
  glutFullScreen();
  // register rendering callback
  glutDisplayFunc(renderscene);
  // register keyboard callback
  glutKeyboardFunc(presskey);
  // specify the clear value for the depth buffer
  glClearDepth(1.0f);
  // set the depth buffering function
  glDepthFunc(GL_LEQUAL);
  // enable depth testing
  glEnable(GL_DEPTH_TEST);
  // Enable face culling: this means that planes will only be rendered when
  // they face towards the camera.
  glEnable(GL_CULL_FACE);
  
  // draws only lines, does not fill them (ie wireframe cube)
  // glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
  
  // draws colors in flat mode (no gradient colors)
  // glShadeModel(GL_FLAT);
  
  // register the rendering timer callback
  glutTimerFunc(1, timerfunc, 1);
  // begin the main glut loop
  glutMainLoop();
}

